#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2022 - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Update the clock information stored in Engine API.

TODO Refactor: This whole class is subject to refactoring, since we are thinking about
     delegating clock information to steering as a provider. This makes especially sense,
     after Steering is providing "virtual timeslots", which currently need to be interpolated
     at multiple places.
TODO Refacotr: It should be reviewed if it's worth storing all the current, additional
     metadata to Engine API.
"""

import logging
from datetime import datetime, timedelta

import aura_engine.scheduling.domain as domain
from aura_engine.base.api import SimpleRestApi
from aura_engine.base.config import AuraConfig
from aura_engine.base.lang import DotDict
from aura_engine.resources import ResourceUtil


class ClockInfoHandler:
    """
    Send current studio clock information to the Engine API endpoint.
    """

    logger = None
    config = None
    api = None
    engine = None

    def __init__(self, engine):
        """
        Initialize.
        """
        self.logger = logging.getLogger("engine")
        self.config = AuraConfig.config()
        self.api = SimpleRestApi()
        self.engine = engine

        if not self.engine:
            self.logger.critical("ClockInfoHandler | Engine instance not available!")
            return

    def on_scheduler_cycle(self, timetable):
        """Call when a new scheduling cycle has been started."""
        # TODO clock updates should also be triggered on scheduler update cycles
        # This will be part of the coming "default station playlist" logic
        pass

    def on_fallback_active(self, timeslot, fallback_name):
        """
        Call when a fallback is activated.
        """
        if timeslot:
            # Only update the current show to the fallback show,
            # in case no timeslot is scheduled
            return

        self.logger.info(f"Fallback '{fallback_name}' activated, clock update required")
        fallback_show_id = self.config.get("fallback_show_id")
        fallback_show_name = self.config.get("fallback_show_name")

        # Interpolate timeslot-less slot
        # TODO start time to be calculated based on previous timeslot (future station logic)
        scheduler = self.engine.scheduler
        upcoming_timeslots: [domain.Timeslot] = scheduler.timetable.get_next_timeslots()
        virtual_start_time = datetime.now()
        virtual_end_time = virtual_start_time + timedelta(hours=1)
        if len(upcoming_timeslots) > 0:
            virtual_end_time = upcoming_timeslots[0].get_start()

        fallback_timeslot = DotDict(
            {
                "timeslotId": -1,
                "timeslotStart": virtual_start_time,
                "timeslotEnd": virtual_end_time,
                "showId": fallback_show_id,
                "showName": fallback_show_name,
                "playlistId": -1,
                "playlistType": -1,
            }
        )
        self.post_clock_info(-1, None, fallback_timeslot, upcoming_timeslots)

    def on_play(self, _):
        """
        Event Handler which is called by the engine when some play command to Liquidsoap is issued.

        This does not indicate that Liquidsoap started playing actually, only that the command has
        been issued. Naturally this happens slightly before single tracks play audio.

        This event is not issued when media is played by Liquidsoap in fallback scenarios.

        Args:
            item (PlaylistItem):

        """
        scheduler = self.engine.scheduler
        active_timeslot: domain.Timeslot = scheduler.timetable.get_current_timeslot()
        active_playlist: domain.Playlist = active_timeslot.get_current_playlist()

        upcoming_timeslots = scheduler.timetable.get_next_timeslots()
        self.post_clock_info(active_playlist, active_timeslot, upcoming_timeslots)

    def post_clock_info(
        self,
        active_playlist: domain.Playlist,
        active_timeslot: domain.Timeslot,
        upcoming_timeslots: [domain.Timeslot],
    ):
        """
        Post current information on timeslots and playlist to the Engine API clock endpoint.
        """
        if len(upcoming_timeslots) >= 2:
            upcoming_timeslots = upcoming_timeslots[0:2]

        built_upcoming = []
        for upcoming_timeslot in upcoming_timeslots:
            built_upcoming.append(self.build_timeslot(upcoming_timeslot, None))

        data = {
            "engineSource": self.config.get("api_engine_number"),
            "currentTimeslot": self.build_timeslot(active_timeslot, active_playlist),
            "upcomingTimeslots": built_upcoming,
            "plannedPlaylist": self.build_playlist(active_playlist),
        }
        url = self.config.get("api_engine_store_clock")
        self.logger.info(f"PUT clock info to '{url}': \n{data}")
        self.api.put(url, data=data)

    def build_timeslot(self, timeslot: domain.Timeslot, playlist: domain.Playlist) -> dict:
        """
        Transform a `Timeslot` object to a dictionary digestable by the clock API endpoint.

        The posted `playlistType` integer indicates if the playlist was assigned on
               (-1) fallback level (no playlist scheduled)
                (0) timeslot level (default)
                (1) schedule level
                (2) show level
                (3) station level

        Args:
            timeslot (Timeslot): The timeslot used for object building.
            playlist (Playlist): The optional playlist used for object building.

        Returns:
            timeslot (dict): JSON representation of the timeslot
        """
        if not timeslot:
            return

        playlist_id = -1
        playlist_type = -1
        if playlist:
            playlist_id = playlist.get_id()
            type = playlist.get_type()
            if type == domain.PlaylistType.TIMESLOT:
                playlist_type = 0
            elif type == domain.PlaylistType.SCHEDULE:
                playlist_type = 1
            elif type == domain.PlaylistType.SHOW:
                playlist_type = 2
            else:
                playlist_type = 3

        show_id = -1
        show_name = None
        if timeslot.show:
            show_id = timeslot.show.id
            show_name = timeslot.show.name

        return {
            "timeslotId": timeslot.get_id(),
            "timeslotStart": timeslot.get_start(),
            "timeslotEnd": timeslot.get_end(),
            "showId": show_id,
            "showName": show_name,
            "playlistId": playlist_id,
            "playlistType": playlist_type,
        }

    def build_playlist(self, playlist: domain.Playlist) -> dict:
        """
        Transform a `Playlist` object to a dictionary digestable by the clock API endpoint.

        Args:
            playlist (Playlist): The playlist domain object.

        Returns:
            ({}): The playlist as dict object for JSON representation.
        """
        if not playlist:
            return None

        items = []
        item: domain.PlaylistItem
        for item in playlist.items:
            content_class = ResourceUtil.get_content_class(item.get_content_type())
            track_type = int(content_class.numeric)
            dict_item = {
                "trackType": track_type,
                "trackStart": item.get_start(),
                "trackNum": item.get_position(),
                "trackDuration": item.duration,
            }
            if item.metadata:
                dict_item |= {
                    "trackArtist": item.metadata.artist,
                    "trackAlbum": item.metadata.album,
                    "trackTitle": item.metadata.title,
                }
            items.append(dict_item)

        pls_dict = {"playlistId": playlist.get_id(), "entries": items}
        return pls_dict
