#
# Aura Engine (https://gitlab.servus.at/aura/engine)
#
# Copyright (C) 2017-now() - The Aura Engine Team.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


"""
Domain models for scheduling.

These domain models are defined:
    - Timeslot: A timetable item.
    - Playlist: Definition of what should be broadcasted for one timeslot.
    - PlaylistType: What kind of playlist it is. Either assigned to timeslot, schedule or show.
    - PlaylistItem: An element of the playlist, referencing media sources and their play state.

"""

from __future__ import annotations

import enum
from typing import NamedTuple

from aura_engine import resources
from aura_engine.base.utils import SimpleUtil
from aura_engine.core.channels import GenericChannel

# from aura_engine.resources import ResourceType, ResourceUtil


class Playlist:
    """
    The playlist assigned to a timeslot.

    Attributes:
        id (int): Playlist ID as used in Steering and Tank.
        timeslot (Timeslot): Back reference to the timeslot.
        type (PlaylistType): Indicating type of timeslot playlist.
        items (PlaylistItem): List of items.

    """

    id: int
    timeslot: Timeslot
    type: PlaylistType
    desc: str
    shuffle: bool
    items: list[PlaylistItem]

    def __init__(self, id: int, desc: str, shuffle: bool = False):
        """
        Initialize.

        Args:
            id (int): The playlist ID.
            desc: (str): Description for the playlist.
            shuffle (bool): If the playlist should be played in random order.
        """
        self.id = id
        self.desc = desc
        self.shuffle = shuffle
        self.items = []
        self.timeslot = None

    def get_id(self) -> int:
        """
        Get the playlist ID.

        Returns:
            int: ID of the playlist.
        """
        return self.id

    def set_timeslot(self, timeslot: Timeslot):
        """
        Set the timeslot.

        Args:
            Timeslot: The timeslot of the playlist.
        """
        self.timeslot = timeslot

    def get_timeslot(self) -> Timeslot:
        """
        Set the timeslot.

        Returns:
            (Timeslot): The timeslot of the playlist.
        """
        return self.timeslot

    def get_type(self) -> PlaylistType:
        """
        Get the playlist type.

        Returns:
            PlaylistType: Type of the playlist.
        """
        return self.type

    def get_items(self) -> list[PlaylistItem]:
        """
        Get the items of the playlist.

        Returns:
            list[PlaylistItem]: All playlist items.
        """
        return self.items

    def get_duration(self) -> float:
        """
        Return the total length of the playlist in seconds.

        Returns:
            float: Length in seconds.

        """
        return sum(item.duration for item in self.items)

    def get_description(self) -> str:
        """
        Get playlist description.

        Returns:
            str: The description.
        """
        return self.desc

    def do_shuffle(self) -> bool:
        """
        Returns true if the playlist should be played in random order.

        Returns:
            bool: `False` if playout mode is linear.

        """
        return self.shuffle

    def add(self, item: PlaylistItem) -> None:
        """
        Add an item to the playlist.

        It also adds the back-reference to the playlist and performs linking to the previous
        playlist item.

        Args:
            item (PlaylistItem): The playlist item to append.

        """
        prev = None
        if self.items:
            prev = self.items[-1]
            item.prev = prev
            item.prev.next = item
        item.playlist = self
        self.items.append(item)

    def __str__(self) -> str:
        """
        String representation.

        Returns:
            str: Playlist as String.

        """
        s = f"ID#{self.id} | {self.get_duration()}seconds"
        if self.desc:
            s += f" ({self.desc})"
        return s


Playlists = NamedTuple(
    "Playlists", [("timeslot", Playlist), ("schedule", Playlist), ("show", Playlist)]
)

Show = NamedTuple(
    "Show",
    [
        ("id", str),
        ("name", str),
        ("type", str),
        ("cat", str),
        ("funding_cat", str),
        ("music_focus", str),
    ],
)

Episode = NamedTuple(
    "Episode",
    [
        ("title", str),
        ("hosts", list[str]),
        ("languages", list[str]),
        ("topics", list[str]),
        ("memo", str),
    ],
)


class Timeslot:
    """
    A timeslot in the programme calendar.

    Beside references to playlists, it holds show and episode metadata.
    Note, some episodes values are inherited from the show,
    some others are overridden by the episode.

    Attributes:
        id (int): Timeslot ID as used in Steering and Tank
        repetition_id (int): In case of a re-broadcast it holds a reference to another timeslot ID.
        start: beginning of timeslot as UNIX epoch.
        end: ending of timeslot as UNIX epoch.

        show.id (int): Show ID as used in Steering and Tank.
        show.name (str): Name of the show.
        show.type (str): The type.
        show.cat (str): The category.
        show.funding_cat (str): The funding category.
        show.music_focus (str): The primary music focus.

        episode.title (str): Title of the episode.
        episode.hosts (list[str]): List of hosts.
        episode.languages (list[str]): List of languages.
        episode.topics (list[str]): List of topics.
        episode.memo: Some notes for the (internal, non-public use only).

        playlist.timeslot (Playlist): Playlist assigned directly to the timeslot.
        playlist.schedule (Playlist): Playlist assigned to the schedule.
        paylist.show (Playlist): Playlist assigned to the show.

    """

    id: int
    repetition_id = int
    start: int
    end: int

    show: Show
    episode: Episode
    playlist: Playlists

    def __init__(
        self, id: int, repetition_id: int, start: float, end: float, show: Show, episode: Episode
    ):
        """
        Initialize.

        Args:
            id (int): ID of the timeslot.
            repetition_id (int): ID of the timeslot this timeslot is a repetition of. None if n/a.
            start (int): Start of the timeslot as UNIX timestamp in seconds.
            end (int): Start of the timeslot as UNIX timestamp in seconds.
            show (Show): Show assigned to the timeslot
            episode (Episode): Episode information representing this broadcast.
        """
        self.id = id
        self.repetition_id = repetition_id
        self.start = start
        self.end = end
        self.show = show
        self.episode = episode
        self.playlist = Playlists(None, None, None)

    def get_id(self) -> int:
        """
        Get ID.

        Returns:
            int: The ID.
        """
        return self.id

    def get_repetition_id(self) -> int:
        """
        Get ID of the timeslot this timeslot is a repetition of.

        Returns:
            int: Repetition timeslot ID. None if it is not a repeating timeslot.
        """
        return self.repetition_id

    def set_playlists(self, pl_timeslot: Playlist, pl_schedule: Playlist, pl_show: Playlist):
        """
        Set all playlists and their respective types at once.

        Args:
            pl_timeslot (Playlist): The timeslot playlist.
            pl_schedule (Playlist): The schedule playlist.
            pl_show (Playlist): The show playlist.

        """
        if pl_timeslot:
            pl_timeslot.type = PlaylistType.TIMESLOT
            pl_timeslot.set_timeslot(self)
        if pl_schedule:
            pl_schedule.type = PlaylistType.SCHEDULE
            pl_schedule.set_timeslot(self)
        if pl_show:
            pl_show.type = PlaylistType.SHOW
            pl_show.set_timeslot(self)

        self.playlist = Playlists(timeslot=pl_timeslot, schedule=pl_schedule, show=pl_show)

    def get_current_playlist(self):
        """
        Retrieve the playlist to be scheduled.

        The playlist to be retrieved is checked in this order:

            1. Timeslot playlist
            2. Schedule playlist
            3. Show playlist

        Returns:
            (Playlist): The playlist currently valid for being scheduled.

        """
        playlist = self.playlist.timeslot
        if not playlist:
            playlist = self.playlist.schedule
            if not playlist:
                playlist = self.playlist.show

        return playlist

    def get_start(self) -> float:
        """
        Get start of timeslot.

        Returns:
            float: start as UNIX epoch timestamp.
        """
        return self.start

    def get_end(self) -> float:
        """
        Get end of timeslot.

        Returns:
            float: end as UNIX epoch timestamp.
        """
        return self.end

    def get_duration(self) -> float:
        """
        Get duration of timeslot in seconds.

        Note, this is not the duration of some actual playlist.

        Returns:
            float: timeslot length in seconds.
        """
        return self.get_end() - self.get_start()

    def get_show(self) -> Show:
        """
        Get show associated with the timeslot.

        Returns:
            Show: The show.
        """
        return self.show

    def get_episode(self) -> Episode:
        """
        Get episode associated with the timeslot.

        Returns:
            Episode: The episode.
        """
        return self.episode

    def __str__(self) -> str:
        """
        String representation.

        Returns:
            str: Timeslot as String.

        """
        start = SimpleUtil.fmt_time(self.start)
        end = SimpleUtil.fmt_time(self.end)
        s = f"ID#{self.id}: {start} - {end}"
        if self.show:
            s += f" [Show#{self.show.id}: {self.show.name}]"
        return s


class PlaylistType(enum.Enum):
    """
    Playlist type defines where it is assigned to the timeslot.

    Attributes:
        TIMESLOT (str): (TIMESLOT) The playlist is assigned to the timeslot itself.
        SCHEDULE (str): (SCHEDULE) The playlist is assigned to the schedule.
        SHOW (str): (SHOW) The playlist is assigned to the show.

    """

    TIMESLOT = "playlist_timeslot"
    SCHEDULE = "playlist_schedule"
    SHOW = "playlist_show"


class PlaylistItem:
    """
    An item of some playlist.

    It holds an reference to the playlist and a double linked list to the previous and next item.

    Attributes:
        prev (PlaylistItem): The previous playlist item.
        next (PlaylistItem): The next playlist item.

        playlist (str): Back reference to playlist.
        source (str): URI referencing the audio source.
        duration (float): Duration in seconds.
        volume (int): Value from 0-100% indicating loudness.

        play_status (PlaylistItem.PlayStatus): Play-out state, set when playing starts.
        play_start_time (int): UNIX epoch, when item actually started playing. \
            Not necessarily matching `start`.
        play_channel (GenericChannel): Play-out channel, set when item is starts playing.
        play_queue_state (QueueState): Progress on queuing, set when item is queued.

    TODO Refactor play and queue state into a single, meaningful object.
         Also using getters and setters.
    """

    prev: PlaylistItem
    next: PlaylistItem

    playlist: Playlist
    source: str
    duration: float
    volume: int

    Metadata = NamedTuple("Metadata", [("artist", str), ("album", str), ("title", str)])
    metadata: Metadata

    class PlayStatus(str, enum.Enum):
        """
        Play-state of a playlist item.
        """

        UNKNOWN = "unknown"
        LOADING = "loading"
        READY = "ready_to_play"
        PLAYING = "playing"
        FINISHED = "finished"

    class QueueState(enum.Enum):
        """
        Types of playlist item behaviours.
        """

        OKAY = "ok"
        CUT = "cut"
        OUT_OF_SCHEDULE = "oos"

    play_status: PlayStatus
    play_start_time: int
    play_channel: GenericChannel
    play_queue_state: QueueState

    def __init__(self, source: str, duration: float, volume: int, metadata: Metadata):
        """
        Initialize the playlist item.

        Args:
            source (str): URI referencing the media source.
            duration (float): Duration in seconds.
            volume (int): Volume of the media source.
            metadata (Metadata): Metadata on the media.
        """
        self.prev = None
        self.next = None

        self.source = source
        self.duration = duration
        self.volume = volume
        self.metadata = metadata

        self.play_status = None
        self.play_start_time = None
        self.play_channel = None
        self.play_queue_state = PlaylistItem.QueueState.OKAY

    def get_playlist(self) -> Playlist:
        """
        Get the playlist of this item.

        Returns:
            Playlist: The associated playlist.
        """
        return self.playlist

    def get_position(self) -> int:
        """
        Calculate the position in the playlist.

        Returns:
            int: number in the list of items.
        """
        if self.prev:
            return 1 + self.prev.get_position()
        return 1

    def get_start(self) -> float:
        """
        Calculate the planned start time of an item as an UNIX epoch timestamp.

        The start is the sum of the timeslot start plus all durations of previous items.

        Returns:
            float: Timestamp when item is expected to start.
        """
        if self.prev:
            return self.get_prev().get_end()
        else:
            return self.playlist.timeslot.start

    def get_end(self) -> float:
        """
        Calculate the end of the item as UNIX epoch timestamp.

        Returns:
            float: Timestamp when item is expected to end.
        """
        return self.get_start() + self.duration

    def get_content_type(self) -> resources.ResourceType:
        """
        Get content type for the given source URI.

        Returns:
            ResourceType: Type of the content.
        """
        return resources.ResourceUtil.get_content_type(self.source)

    def get_prev(self) -> PlaylistItem:
        """
        Get previous playlist item.

        Returns:
            PlaylistItem: The previous item.
        """
        return self.prev

    def get_next(self) -> PlaylistItem:
        """
        Get next playlist item.

        Returns:
            PlaylistItem: The next item.
        """
        return self.next

    def get_all_prev(self) -> list[PlaylistItem]:
        """
        Retrieve all previous items as part of the current item's playlist.

        Returns:
            list[PlaylistItem]: All previous items.
        """
        prev = self.get_prev()
        if prev:
            all_prev = prev.get_all_prev()
            all_prev.append(prev)
            return all_prev
        return []

    def get_all_next(self, timeslot_sensitive: bool = True) -> list[PlaylistItem]:
        """
        Retrieve all following items as part of the current item's playlist.

        Args:
            timeslot_sensitive (Boolean): If `True` items which start after \
                the end of the timeslot are excluded.

        Returns:
            list[PlaylistItem]: List of PlaylistItem.

        """
        next = self.get_next()

        if next:
            if timeslot_sensitive:
                if next.get_start() > self.playlist.timeslot.end:
                    return []
            all_next = next.get_all_next(timeslot_sensitive)
            all_next.append(self)
            return all_next
        else:
            return []

    def remove(self):
        """
        Delete the item from the playlist.
        """
        prev = self.prev
        if self.prev:
            self.prev.next = self.next
        if self.next:
            self.next.prev = prev
        self.playlist.items.remove(self)
        self.playlist = None
        self.prev = None
        self.next = None

    def set_play_status(self, status: PlayStatus):
        """
        Set the play-out status of the current item.

        Args:
            status (PlayStatus): Info on the play status.
        """
        self.play_status = status

    def get_play_status(self):
        """
        Returns the play-out status.

        Returns:
            PlayStatus: Info on the items is being played, buffered etc.
        """
        return self.play_status

    def __str__(self):
        """
        String representation.

        Returns:
            str: Playlist item as String.

        """
        start = SimpleUtil.fmt_time(self.get_start())
        end = SimpleUtil.fmt_time(self.get_end())
        track = self.source[-25:]

        return f"PlaylistItem [{start} - {end} " + f"| {self.duration}sec | Source: ...{track}]"
